      PROGRAM D2O
C                                               Version 2.0, September 1987
C  'D2O' evaluates density, internal energy, enthalpy and entropy for
C   heavy water in kPa - KJ - K units.  Subcooled, superheated and saturation
C   properties are evaluated for temperatures from the triple point to the
C   critical point.
C   
C     Ref.: Tables of Thermodynamic Properties of Heavy Water in SI Units,
C     (Prepared for Atomic Energy of Canada Limited and Ontario Hydro
C     by P.G. Hill, R.D. MacMillan, and Victor Lee, University of B.C.)
C     AECL 7531, Dec. 1981, pg. 169, equations 1 thru 7.
C     Note error on page 167:  Tc = 1.553, not 1
C
C   
C   Take note that a significant range of the saturation values Vf,Vg,Hf,
C   Hg are quick high-ordered curve fits for economy's sake. See local comments
C   in the subroutines, for more details.  The accuracy of these fits are
C   within .01% of Hill's equations, which could readily be substituted.
C
C   There are two constants (C1 and C2 on page 167 of the reference) which
C   could be fine-tuned for accurate 5th/6th significant digits.  This has
C   not been done, as it depends on the machine being used.  This code
C   has been run on a VAX 8600 mini, and a Multitech Plus 700 PC (compiled
C   with VAX Fortran V4.0, and Microsoft Fortran-77 V3.3 respectively).
C   There seem to be no significant discrepancies between the
C   values produced by this formulation and those of the tables of
C   the reference, for the range from triple point to critical point.
C   The author would appreciate being notified of any that are found.
C
C   The author and Ontario Hydro make no warranties expressed or implied
C   as to the performance of this program.  Neither the author nor
C   Ontario Hydro will be held liable for any direct, indirect, incidental,
C   or consequential damages resulting from the use of this program.
C   Your use of the program consitutes your agreement to this disclaimer
C   and your release of the author and Ontario Hydro from any form of
C   liability or litigation.  This is not an official Ontario Hydro release;
C   it is under development and is hereby available for trial use.  You
C   may copy and modify this program provided this disclaimer remains intact.
C
C   Direct queries to:
C                              Bob Horvath
C                              Nuclear Systems Department
C                              Ontario Hydro, H10B10
C                              700 University Ave,           
C                              Toronto,  M5G 1X6
C                              (416) 592-3792
C                       
      CHARACTER*1 A,AB,TP
C  NDP = number of decimal places in input/output of pressure & temp.
      DATA NDP/2/
C  Default values for question responses
      AB='A'
      TP='T'
 100  WRITE(*,10) AB
  10  FORMAT(/,10X,'D2O PROPERTIES'//,6X,'A. Saturation tables'/6X
     +,'B. Superheated/subcooled table',//,' Select (A/B): [',A1,'] ',$)
      READ(*,'(A1)') A
C
C  Saturation properties ###############################################
C
      IF (A.EQ.'A'.OR.A.EQ.'a'.OR.(A.EQ.' '.AND.AB.EQ.'A')) THEN
        AB='A'
        WRITE(*,15) TP
  15    FORMAT(' Temperature or Pressure-based (T/P): [',A1,'] ',$)
        READ(*,'(A1)') A
C  Temperature tables
        IF (A.EQ.'T'.OR.A.EQ.'t'.OR.A.EQ.' '.AND.TP.EQ.'T') THEN
          TP='T'
          CALL LOOP('Temp. (deg.C)',ISTART,ISTOP,ISTEP,NDP)
          WRITE(*,20)              
  20      FORMAT(//,20X,'SATURATION TABLE FOR D2O BASED ON TEMPERATURE'/
     + /,'  DEG.C   Psat(KPa)      Vf (m3/Kg)  Vg        Hf (KJ/Kg) Hg'
     + ,'      Uf (KJ/Kg) Ug')
          DO 200 I=ISTART,ISTOP,ISTEP
            T=I/10.**NDP + 273.15
            PS=PSD2O(T)
C  Calling HF right after VF economizes in RHOD2O call, i.e. calling
C  Vf,Vg,Hf,Hg is less efficient than Vf,Hf,Vg,Hg.  No direct call
C  for Uf & Ug is available, just use U=H-P*V.
            VF=VFD2O(T)
            HF=HFD2O(T)
            UF=HF-PS*VF
            VG=VGD2O(T)
            HG=HGD2O(T)
            UG=HG-PS*VG
            WRITE(*,30) T-273.15,PS,VF,VG,HF,HG,UF,UG
  30        FORMAT(1X,F7.2,1X,F9.3,3X,F9.7,F12.6,2(3X,F7.2,3X,F6.1))
 200      CONTINUE
C  Pressure Tables
        ELSEIF (A.EQ.'P'.OR.A.EQ.'p'.OR.A.EQ.' '.AND.TP.EQ.'P') THEN
          TP='P'             
          CALL LOOP('Pressure(kPa)',ISTART,ISTOP,ISTEP,NDP)
          WRITE(*,40)
  40      FORMAT(//,20X,'SATURATION TABLE FOR D2O BASED ON PRESSURE',//
     + ,'  P(KPa)   Tsat (C)      Vf (m3/Kg)  Vg        Hf (KJ/Kg) Hg'
     + ,'      Uf (KJ/Kg) Ug')
          DO 300 I=ISTART,ISTOP,ISTEP
            PS=I/10.**NDP   
            T=TSD2O(PS)
C  Calling HF immediately after VF economizes in RHOD2O call
            VF=VFD2O(T)
            HF=HFD2O(T)
            UF=HF-PS*VF
            VG=VGD2O(T)
            HG=HGD2O(T)
            UG=HG-PS*VG
            WRITE(*,50) PS,T-273.15,VF,VG,HF,HG,UF,UG
  50        FORMAT(1X,F7.2,1X,F9.3,3X,F9.7,F12.6,2(3X,F7.2,3X,F6.1))
 300      CONTINUE
        END IF
C
C  Superheated or subcooled properties #################################
C
      ELSEIF (A.EQ.'B'.OR.A.EQ.'b'.OR.A.EQ.' '.AND.AB.EQ.'B') THEN
        AB='B'
        WRITE(*,'(/,'' Enter pressure (KPa): '',$)')
        READ(*,*) P                             
        CALL LOOP('Temp. (deg.C)',ISTART,ISTOP,ISTEP,NDP)
        WRITE(*,60) P,TSD2O(P)-273.15
  60    FORMAT(//,6X,'LIQUID AND VAPOUR TABLE OF PROPERTIES FOR D2O',/
     +  /,' PRESSURE =',F9.2,' KPa,    SAT. TEMP. =',F7.2,' DEG C',/
     +  ,'  DEG.C   V, m3/Kg      H, KJ/Kg    U, KJ/Kg   S, KJ/(KgK)')
C
        DO 400 I=ISTART,ISTOP,ISTEP
          T=I/10.**NDP+273.15
          V=1./RHOD2O(P,T)
          U=UD2O(P,T)
          H=HD2O(P,T)
          S=SD2O(P,T)
          WRITE (*,'(F7.2,F12.7,2F12.1,F12.4)') T-273.15,V,H,U,S
 400    CONTINUE
      END IF
C
C
      WRITE(*,'(/,'' Another (Y/N)? [Y] '',$)')
      READ(*,'(A1)') A        
      IF (A.EQ.'Y'.OR.A.EQ.'y'.OR.A.EQ.' ') GO TO 100
      STOP
      END
C
C***********************************************************************
      SUBROUTINE LOOP(TYPE,ISTART,ISTOP,ISTEP,NDP)
C  Get temp/press range & assign DO-loop indices. NDP = no. of decimals
      CHARACTER*13 TYPE
      WRITE(*,'(1X,''Initial '',A13,'': '',$)') TYPE
      READ(*,*) START
      WRITE(*,'(1X,''  Final '',A13,'': '',$)') TYPE
      READ(*,'(F9.0)') STOP
      WRITE(*,'(1X,''   Step '',A13,'': '',$)') TYPE
      READ(*,'(F9.0)') STEP
      IF (STOP.EQ.0) STOP=START
      IF (STEP.EQ.0.) STEP=STOP-START
      ISTART=START*10**NDP
      ISTOP = STOP*10**NDP
      ISTEP = STEP*10**NDP
      IF (ISTEP.EQ.0) ISTEP=1
      RETURN
      END
C***********************************************************************
      FUNCTION PSD2O(T)
C  Calculates saturation pressure (KPa) given temperature T in deg. K.
C     Ref.: Tables of Thermodynamic Properties of Heavy Water in SI Units,
C     AECL 7531, Dec. 1981, pg. 171 eqn. 15.
C
      COMMON /SAT/TAU,ZZ
      TC=643.89
C  Check range, LOG(0.) error gives traceback info
      IF (T.LT.276.949.OR.T.GT.643.8901) THEN
        WRITE (*,10) T
  10    FORMAT(/' *** TEMP.',F9.2,' DEG. K OUT OF RANGE IN PSD2O ***'/)
        ZZ=LOG(0.)
      END IF
      TAU=1.-T/TC
      ZZ = TC/T*( +4.19174*TAU**10 -3.92488*TAU**5.5 -18.1747*TAU*TAU
     2                        +17.6012*TAU**1.9 -7.81583*TAU)
      PSD2O = 21660.*EXP(ZZ)
      RETURN
      END
C***********************************************************************
      FUNCTION TSD2O(P)
C
C  Calculates saturation temperature (deg.K) given pressure in KPa,
C  using Function PSD20(T) and Newton's Method for finding roots
C     Ref.: Tables of Thermodynamic Properties of Heavy Water in SI Units,
C     AECL 7531, Dec. 1981, pg. 171 eqn. 15.
C
      COMMON /SAT/TAU,ZZ
C  Initial guess based on eqn. 15 with linear term only, adjusted slightly.
C  This guess gives rel.diff.<0.00005 within 3 iterations for all P<Pc.
      TSD2O=643.89/(1.-LOG(P/21660)/7.85)
      DO 100 I=1,3
        PS=PSD2O(TSD2O)
        F=PS-P
        FPRIME= -PS/TSD2O* (ZZ +41.9174*TAU**9 -21.5868*TAU**4.5
     +                   -36.3494*TAU +33.4423*TAU**0.9 -7.81583)
        TSD2O = TSD2O - F/FPRIME
  100 CONTINUE
      RETURN
      END
C***********************************************************************
      FUNCTION VFD2O(T)
C  Returns specific volume of saturated liquid in m3/Kg.  T in deg. K
      REAL*8 TT
      IF (T.GT.280. .AND. T.LT.591.) THEN
C  Curve fit for above range is within 0.01% of Hill's equations
        TT=T*1.D-2
        VFD2O = EXP( -6.5935639 +1.21023444*TT -2.1610343*TT*TT
     2 +1.3311899*TT**3 -.4200469*TT**4 +7.3678383D-2*TT**5
     3  - 6.854802D-3*TT**6 + 2.6541181D-4*TT**7 )
      ELSE
C  If close to triple or critical points, use Hill's fits (via RHO)
        PS=PSD2O(T)*1.00001
        VFD2O=1.0/RHOD2O(PS,T)
      END IF
      RETURN
      END
C***********************************************************************
      FUNCTION VGD2O(T)
      REAL*8 TT
C  Returns specific volume of saturated gas in m3/Kg.  T in deg. K
      IF (T.LT.591.) THEN
C  Curve fit for above range is within 0.01% of Hill's equations
        TT=T*1.D-2
        VGD2O =EXP( 112.782723D0 - 121.099804D0*TT + 62.97936106D0*TT*TT
     2  - 20.0516607D0*TT**3 + 4.06673812D0*TT**4 - .515308316D0*TT**5
     3  + 3.7312542127D-2*TT**6 - 1.1824863609D-3*TT**7 )
      ELSE
C  If close to triple point, use Hill's fits (thru RHO)
        PS=PSD2O(T)*.999999
        VGD2O=1.0/RHOD2O(PS,T)
      END IF
      RETURN
      END
C***********************************************************************
      FUNCTION HFD2O(T)
      REAL*8 TT
C  Calculates internal energy of saturated liquid D2O in KJ/Kg. T in K.
      IF (T.LT.591.) THEN
C  Curve fit for above range is within 0.02 KJ/Kg of Hill's equations
        TT=T*1.D-2         
        HFD2O =  8549.658 - 18180.2334D0*TT + 15535.561D0*TT*TT
     2   - 7416.07913D0*TT**3 + 2218.0784D0*TT**4 - 425.9972378D0*TT**5
     3  +51.283009364D0*TT**6 -3.534432925D0*TT**7 +.10672246654D0*TT**8
      ELSE
C  If close to critical point, use Hill's fits (via RHO)
        PS=PSD2O(T)*1.00001
        RHO=RHOD2O(PS,T)
        HFD2O=U0D2O(RHO,T) + PS/RHO
      END IF
      RETURN            
      END 
C***********************************************************************
      FUNCTION HGD2O(T)
      REAL*8 TT
C  Calculates internal energy of saturated liquid D2O in KJ/Kg. T in K.
      IF (T.LT.591.) THEN
C  Curve fit for above range is within 0.004% of Hill's equations
        TT=T*1.D-2
        HGD2O =  7595.171 - 10439.8385D0*TT + 8358.0858D0*TT*TT
     2   - 3635.46124D0*TT**3 + 941.558218D0*TT**4 - 144.908137D0*TT**5
     3   + 12.2477167D0*TT**6 - .4393428905D0*TT**7 
      ELSE
C  If close to critical point, use Hill's fits (via RHO)
        PS=PSD2O(T)*.999999
        RHO=RHOD2O(PS,T)
        HGD2O=U0D2O(RHO,T) + PS/RHO
      END IF
      RETURN
      END 
C***********************************************************************
      FUNCTION HD2O(P,T)
C  Calculates enthalpy of  D2O in KJ/Kg. P in KPa, T in K.
C  Determine saturation density (Kg/m3) for U0D2O call
      RHO=RHOD2O(P,T)
C
      HD2O=U0D2O(RHO,T) + P/RHO
      RETURN
      END
C***********************************************************************
      FUNCTION UD2O(P,T)
C
C  Calculates internal energy in KJ/Kg given p in KPa and T in deg. K
C  Determine density in Kg/m3 for given P and T
      RHO=RHOD2O(P,T)
C
      UD2O=U0D2O(RHO,T)
      RETURN
      END
C***********************************************************************
      FUNCTION U0D2O(RHO0,T0)
C  Calculates internal energy in KJ/Kg given p in KPa and T in deg. K
C     Ref.: Tables of Thermodynamic Properties of Heavy Water in SI Units,
C     AECL 7531, Dec. 1981, pg. 169 eqn before eqn. 5
C
      REAL*8 RHO,T,SI,SI0,SIPLUS,DT,SMALL,DSIDT
C  Economize if same RHO and T as previous call
      SAVE RHO,T,UD2O
      IF (REAL(RHO).EQ.RHO0.AND.REAL(T).EQ.T0) THEN
        U0D2O=UD2O
        RETURN
      END IF
      RHO=DBLE(RHO0)
      T=DBLE(T0)
C  Forward difference for 1st derivative of SI w.r.t T (all REAL*8)
      SMALL=1.D-6
      DT=T*SMALL
      SIPLUS=SI(RHO,T+DT)
      SI0   =SI(RHO,T)
      DSIDT = (SIPLUS - SI0)/DT
C  Chain and product rules applied to derivative for U
      UD2O = REAL(SI0 - DSIDT*T)
      U0D2O=UD2O
      RETURN
      END
C***********************************************************************
      FUNCTION SD2O(P0,T0)
C
C  Calculates entropy in KJ/(Kg-K) given p in KPa and T in deg. K
C     Ref.: Tables of Thermodynamic Properties of Heavy Water in SI Units,
C     AECL 7531, Dec. 1981, pg. 169 eqn before eqn. 6
C
      REAL*8 RHO,T,SI,SI0,SIPLUS,DT,SMALL,DSIDT
C  Economize if same P and T as previous call               
      SAVE P,T,DSIDT
      IF (P.EQ.P0.AND.REAL(T).EQ.T0) THEN
        SD2O=REAL(-DSIDT)
        RETURN
      END IF
      P=P0
      T=DBLE(T0)
C  Determine density in Kg/m3 for given P0 and T0
      RHO=DBLE(RHOD2O(P0,T0))
C  Forward difference for 1st derivative of SI w.r.t T (all REAL*8)
      SMALL=1.D-6
      DT=T*SMALL
      SIPLUS=SI(RHO,T+DT)
      SI0   =SI(RHO,T)
      DSIDT = (SIPLUS - SI0)/DT
      SD2O = REAL(-DSIDT)
      RETURN
      END
C***********************************************************************
      FUNCTION RHOD2O(P0,T0)
C
C  Calculates density in Kg/m3 given P0 in KPa and T0 in deg. K (REAL*4)
C     Ref.: Tables of Thermodynamic Properties of Heavy Water in SI Units,
C     AECL 7531, Dec. 1981, pg. 167 eqn before eqn. 4
C
      REAL*8 SI,SIMNUS,SIMID,SIPLUS,DSIDR,D2SIDR,RHO,DRHO,SMALL,T
      COMMON RGAS
C  Economize if same P and T as previous call
      SAVE P,T,RHO
      IF (P.EQ.P0.AND.REAL(T).EQ.T0) THEN
        RHOD2O=REAL(RHO)
        RETURN
      END IF
      P=P0
      T=DBLE(T0)
C                      
C  RHO guess in Kg/m3; valid for ranges below
C  SQRT(-1) gives error exit with trace back info if out of this range
      IF (T0.GT.1073.16.OR.P0.GT.21660.1) THEN
        WRITE(*,'(/'' *** NO GUESS IN RHOD2O SUBR ***  P(KPa),T(K)='')')
        WRITE(*,*) P0,T0
        P=SQRT(-1.)
      END IF
C  Following is guess for liquid, a rough fit of saturation values
      IF (T0.LE.643.889.AND.P0.GT.PSD2O(T0)) THEN
        RHO =EXP( .7955953 + .0614442*T0 - 2.222647E-4*T0*T0
     2            + .3501397E-6*T0**3 - 2.062136E-10*T0**4)
      ELSE
C  For gas phase guess, use Dieterici eqn. of state
C  ref. THERMODYNAMICS (1st ed) by E.F. Obert, eqn. 7-19 pg. 195
C  Solved with Newton's method.  Valid for all subcritical, some supercr
        A=4.*(RGAS*643.89/2.718282)**2/21660.
        B=RGAS*643.19/(21660.*2.718282**2)
        V=RGAS*T0/P
 100    V0=V
        EX=EXP(-A/(RGAS*T0*V0))
        F=RGAS*T*EX-P*(V-B)
        FPRIME=A*EX/(V*V)-P
        V=V0-F/FPRIME
        IF (ABS(V-V0)/V.GT..01) GO TO 100
        RHO=1.D0/DBLE(V)
      END IF
C
C  Calculation of actual RHO, in Kg/m3
C
      SMALL=1.D-5
 200  RHO0=REAL(RHO)
        DRHO=RHO*SMALL
        SIPLUS=SI(RHO+DRHO,T)
        SIMID =SI(RHO,T)
        SIMNUS=SI(RHO-DRHO,T)
C  Central differences for 1st and 2nd derivatives
        DSIDR = (SIPLUS - SIMNUS)/(2.D0*DRHO)
        D2SIDR= (SIPLUS -2.D0*SIMID + SIMNUS)/(DRHO*DRHO)
C  Newton's Method
        F = REAL(RHO*RHO*DSIDR) - P
        FPRIME = REAL(RHO*(2.D0*DSIDR + RHO*D2SIDR))
        RHO = DBLE(RHO0 - F/FPRIME)
      IF (ABS((REAL(RHO)-RHO0)/RHO0).GT.1.E-5) GO TO 200
      RHOD2O=REAL(RHO)
      RETURN
      END
C***********************************************************************
      FUNCTION SI(RHO,T)
C
C  SI = Hemholtz function, KJ/(KG K), RHO in Kg/m3, T in deg. K
C   Ref.: Tables of Thermodynamic Properties of Heavy Water in SI Units,
C   AECL 7531, Dec. 1981, pg. 166 eqns 1,2,3
C
      IMPLICIT REAL*8 (A-H,O-Z)
      REAL*8 ACOEF(10,7),CCOEF(8),LNT
      REAL*4 RGAS
      COMMON /COEF/ACOEF,CCOEF
      COMMON RGAS
C  Economize if same temperature as last call
      SAVE T0,SI0
      TAU=1.D+3/T
      IF (T.EQ.T0) GO TO 150
        T0=T
        TAUINV=T/1.D+3
C
C  Evaluate SIo(T)
        SI0=0.D0
        DO 100 I=6,2,-1
          SI0 = SI0 + CCOEF(I)*TAUINV**(I-1)
 100    CONTINUE
        LNT = DLOG(T)
        SI0 = SI0 + CCOEF(8)*TAUINV*LNT + CCOEF(1) + CCOEF(7)*LNT
C
C  Evaluate Q function, with density in g/cm3 as required by fit.
C  
 150  RHOGCM=RHO*1.D-3
      SUMJ=0.D0
      DO 300 J=1,7
        TAUAJ=2.53D0
        RHOAJ=1.1D0
        IF (J.EQ.1) THEN
          TAUAJ=1.553D0
          RHOAJ=0.7D0
        END IF
        SUMI=ACOEF(1,J)
        DO 200 I=2,8
          IF (I.GT.4.AND.J.GT.2) GO TO 250
          SUMI = SUMI + ACOEF(I,J)*(RHOGCM-RHOAJ)**(I-1)
 200    CONTINUE
 250    SUMI = SUMI +EXP(-4.3D0*RHOGCM)*(ACOEF(9,J) +ACOEF(10,J)*RHOGCM)
        SUMJ = SUMJ + SUMI*(TAU-TAUAJ)**(J-2)
 300  CONTINUE
      Q = (TAU-1.553D0)*SUMJ
      SI = SI0 + DBLE(RGAS)*T*( DLOG(RHOGCM) + RHOGCM*Q )
      RETURN
      END
C***********************************************************************
      BLOCK DATA CONST
C  Assigns Aij coefficients
C  Ref.:  Tables of Thermodynamic Properties of Heavy Water in SI Units,
C         AECL 7531, Dec. 1981, pg. 168, Table 1.
C
      REAL*8 ACOEF(10,7),CCOEF(8)
      COMMON /COEF/ACOEF,CCOEF
      COMMON RGAS
      DATA ACOEF/73.13848592, -285.20415917, 535.71659288,-649.81000614,
     +574.63280680,-387.92157774,206.34569512,-79.89428513,-996.36169097
     +,-766.27290006,24.74108348,-105.57317181,200.87302906,-235.1877644
     +,224.56976938,-40.09924297,128.77154771,-28.40907978,
     +-1389.08003142,-1672.09705556,11.64775625,-42.51820251,72.45541064
     +,-82.55391089,4*0.,-267.8548252,-998.6498271,2.66566642,
     +-9.19657655,15.1309692,-7.24860975,4*0.,-46.8390432,-227.34793319,
     +-6.73408249,24.03602093,-41.0807983,45.39111005,4*0.,139.21659329,
     +566.02305152,-5.24802962,18.52690633,-31.42397369,26.43208802,4*0.
     +,96.31411481,453.20280933,-1.17583447,4.13816432,-6.55842224
     +,4.75774631,4*0.,19.39184297,103.56819758/
      DATA CCOEF/ 1866.73, 4661.9, 64.605, -284.8833, 100.1333, -13.135,
     2        0.32684, -1211.253/
      DATA RGAS/0.41515/
      END

